package edu.washington.cs.radishsim;

/** the type of in-hardware status bits */
enum InHardwareStatus {
   // IN_SOFTWARE status for a location l is represented by leaving l unmapped
   EVERYTHING, ALL_LEAF_READS, LAST_WRITE;
}

enum MemoryAccessType {
   READ, WRITE
}

enum SynchronizationOperationType {
   SOURCE, SINK
}

class MemorySystemConstants {

   public static final int L1_HIT_LATENCY            = 1;
   public static final int L2_HIT_LATENCY            = 10;
   public static final int REMOTE_HIT_LATENCY        = 15;
   public static final int L3_HIT_LATENCY            = 35;
   public static final int MEMORY_ACCESS_LATENCY     = 120;

   // TODO: make a more refined estimate here
   public static final int APPROX_SW_CHECK_COST      = 100;
   public static final int FASTTRACK_CHECK_INSN_COST = 10;

   /** How many bytes of md we have per byte of data in the *expanded* encoding. */
   public static short EXPANDED_METADATA_BYTES_PER_DATA_BYTE() {
      return 2;
   }

   /** How many bytes of md we have per byte of data in the *compact* encoding. */
   public static short COMPACT_METADATA_BYTES_PER_DATA_BYTE() {
      return 1;
   }

   private static boolean validLineSize = false;
   private static int     lineSize;
   private static int     lineSizeBits;

   /** NB: must be called before any GenericAccess objects are created. */
   public static void setLineSize(int ls) {
      assert !validLineSize;
      unsafeSetLineSize(ls);
   }

   /**
    * Ordinarily, the line size should be set only once per execution; setLineSize() enforces this
    * invariant. This unsafe version allows the line size to be reset, and should be used for
    * testing purposes only.
    */
   static void unsafeSetLineSize(int ls) {
      lineSize = ls;
      lineSizeBits = BitTwiddle.floorLog2(ls);
      validLineSize = true;
   }

   /**
    * The cache line size used throughout this system. NB: this simulator only supports a constant
    * global line size.
    */
   public static int LINE_SIZE() {
      assert validLineSize;
      return lineSize;
   }

   /** The bitmask used to mask out everything but the offset within a line. */
   public static long LINE_OFFSET_MASK() {
      assert validLineSize;
      return lineSize - 1;
   }

   /** The number of bits needed to represent the line size. */
   public static int LINE_SIZE_BITS() {
      assert validLineSize;
      return lineSizeBits;
   }

   /**
    * Used to keep compact md lines aligned to double-line boundaries. It's a double-line because of
    * md scaling.
    */
   public static boolean COMPACT_MD_ALIGNED(long a) {
      assert validLineSize;
      return 0 == (a & COMPACT_MD_LINE_OFFSET_MASK());
   }

   /** Used to align addresses to the compact-md-alignment requirements. */
   public static long COMPACT_MD_ALIGNMENT_MASK() {
      return ~COMPACT_MD_LINE_OFFSET_MASK();
   }

   /** The bitmask used to mask out everything but the offset within a compact md line. */
   public static long COMPACT_MD_LINE_OFFSET_MASK() {
      assert validLineSize;
      return (lineSize * 2) - 1;
   }

   public static boolean COMPACT_MD_TOP(long a) {
      long offset = a & COMPACT_MD_LINE_OFFSET_MASK();
      return offset >= DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted.COMPACT);
   }

   public static boolean COMPACT_MD_BOTTOM(long a) {
      long offset = a & COMPACT_MD_LINE_OFFSET_MASK();
      return offset < DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted.COMPACT);
   }

   /** The number of data bytes "covered" by 1 line's worth of metadata */
   public static int DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted c) {
      assert validLineSize;

      switch ( c ) {
      case COMPACT:
         return lineSize / COMPACT_METADATA_BYTES_PER_DATA_BYTE();
      case EXPANDED:
         return lineSize / EXPANDED_METADATA_BYTES_PER_DATA_BYTE();
      }

      throw new IllegalArgumentException();
   }

   public static long SOFTWARE_MD_ADDR_OF_HW_MD(UnownedMetadataByteAddress umba) {
      return 0x100000 + (umba.get() * 4);
   }

   public static long SOFTWARE_MD_VC(short tid) {
      return 0x200000 + (tid * 4);
   }

}

abstract class GenericAccess {
   /** cpuid used when manipulating "global" metadata addresses */
   static final CpuId         GLOBAL_CPUID = new CpuId(0);
   /** cpuid used when manipulating data addresses */
   static final CpuId         DATA_CPUID   = new CpuId(0);

   protected MemoryAccessType type;
   protected ByteAddress      baseAddr;
   protected ByteAddress      addr;
   protected int              size;

   abstract void incr();

   abstract boolean keepGoing();

   final boolean write() {
      return type == MemoryAccessType.WRITE;
   }

   final boolean read() {
      return type == MemoryAccessType.READ;
   }

   final MemoryAccessType type() {
      return type;
   }

   final int size() {
      return size;
   }

   ByteAddress addr() {
      return addr;
   }

   LineAddress lineAddress() {
      return addr.lineAddress();
   }

   short lineOffset() {
      return addr.lineOffset();
   }

   @Override
   public String toString() {
      return "type:" + String.valueOf(this.type) + " base:" + this.baseAddr.toString() + " addr:"
             + this.addr.toString() + " size:" + this.size;
   }
}

class MetadataAccess extends GenericAccess {

   protected void checkExpanded() {
      // ensure access fits within an expanded metadata cache line
      int lastByte = lineOffset()
                     + (size * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      assert lastByte <= MemorySystemConstants.LINE_SIZE();
   }

   /**
    * @param mtype
    *           whether the access is a read or write
    * @param addr
    *           the metadata address of the first byte in the access
    * @param size
    *           size of the access (in data bytes)
    */
   public MetadataAccess(MemoryAccessType mtype, MetadataByteAddress addr, final int size) {
      this.type = mtype;
      // make defensive copies
      this.baseAddr = new MetadataByteAddress(addr);
      this.addr = new MetadataByteAddress(addr);
      this.size = size;
      checkExpanded();
   }

   /** Copy constructor */
   public MetadataAccess(final MetadataAccess access) {
      this.type = access.type();
      // make defensive copies
      this.baseAddr = new MetadataByteAddress((MetadataByteAddress) access.baseAddr);
      this.addr = new MetadataByteAddress(access.addr());
      this.size = access.size();
      checkExpanded();
   }

   protected void checkCompact() {
      // ensure access fits within an expanded metadata cache line
      int lastByte = lineOffset()
                     + (size * MemorySystemConstants.COMPACT_METADATA_BYTES_PER_DATA_BYTE());
      assert lastByte <= MemorySystemConstants.LINE_SIZE();
   }

   /**
    * Used for iterating over a md cache line. Hence, there's no access type and there's
    * less-restrictive bounds checking.
    */
   public MetadataAccess(final MetadataByteAddress mba, final int size) {
      this.type = null;
      // make defensive copies
      this.baseAddr = new MetadataByteAddress(mba);
      this.addr = new MetadataByteAddress(mba);
      this.size = size;
      checkCompact();
   }

   @Override
   public MetadataByteAddress addr() {
      return (MetadataByteAddress) this.addr;
   }

   @Override
   public MetadataLineAddress lineAddress() {
      return ((MetadataByteAddress) this.addr).lineAddress();
   }

   public short metadataLineOffset(RadishLine.Compacted c) {
      return this.addr().metadataLineOffset(c);
   }

   public CpuId owner() {
      return ((MetadataByteAddress) this.addr).owner();
      /*
       * long o = ((addr >> (13 * 4)) & 0x0F); assert (o != 15); return new CpuId((int) o);
       */
   }

   @Override
   public void incr() {
      assert keepGoing();
      addr.incr();
   }

   @Override
   boolean keepGoing() {
      return addr.get() < (baseAddr.get() + (size * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE()));
   }

   /** @return true if the given byte is within this access */
   public boolean contains(final MetadataByteAddress byteAddr) {
      final long b = byteAddr.get();
      final long limit = baseAddr.get()
                         + (size * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      return (b >= baseAddr.get() && b < limit);
   }

} // end class MetadataAccess

class UnownedMetadataAccess extends GenericAccess {

   protected void check() {
      // ensure access fits within an expanded metadata cache line
      int lastByte = lineOffset()
                     + (size * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      assert lastByte <= MemorySystemConstants.LINE_SIZE();
   }

   /** Copy constructor from an owned metadata access */
   public UnownedMetadataAccess(final MetadataAccess access) {
      this.type = access.type();
      this.baseAddr = new UnownedMetadataByteAddress((MetadataByteAddress) access.baseAddr);
      this.addr = new UnownedMetadataByteAddress(access.addr());
      this.size = access.size();
      check();
      // this(access, access.size());
   }

   /** Copy constructor from an unowned metadata access */
   public UnownedMetadataAccess(final UnownedMetadataAccess access) {
      this.type = access.type();
      this.baseAddr = new UnownedMetadataByteAddress((UnownedMetadataByteAddress) access.baseAddr);
      this.addr = new UnownedMetadataByteAddress(access.addr());
      this.size = access.size();
      check();
   }

   @Override
   public UnownedMetadataByteAddress addr() {
      return (UnownedMetadataByteAddress) this.addr;
   }

   @Override
   public UnownedMetadataLineAddress lineAddress() {
      return ((UnownedMetadataByteAddress) this.addr).lineAddress();
   }

   public short metadataLineOffset(RadishLine.Compacted c) {
      return this.addr().metadataLineOffset(c);
   }

   @Override
   public void incr() {
      assert keepGoing();
      this.addr.incr();
   }

   @Override
   boolean keepGoing() {
      return this.addr.get() < (this.baseAddr.get() + (this.size * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE()));
   }

} // end class MetadataAccess

class DataAccess extends GenericAccess {
   DataAccess(MemoryAccessType mtype, ByteAddress addr, final int size) {
      this.type = mtype;
      // make defensive copies
      this.baseAddr = new DataByteAddress(addr.get());
      this.addr = new DataByteAddress(addr.get());
      this.size = size;
      // ensure access fits within a cache line
      assert (lineOffset() + size <= MemorySystemConstants.LINE_SIZE());
   }

   @Override
   public DataByteAddress addr() {
      return (DataByteAddress) this.addr;
   }

   @Override
   void incr() {
      assert keepGoing();
      addr.incr();
   }

   @Override
   boolean keepGoing() {
      return addr.get() < (baseAddr.get() + size);
   }
}
