package edu.washington.cs.radishsim;

import java.util.Arrays;
import java.util.Map;

import edu.washington.cs.radishsim.RadishLine.ByteMetadata;
import static edu.washington.cs.radishsim.MemorySystemConstants.*;

/** the type of local permissions */
enum MemoryAccessPermission {
   NOTHING_OK, READ_OK, WRITE_OK;

   public static final Comparison<MemoryAccessPermission> Compare = new Comparison<MemoryAccessPermission>();
}

abstract class VCPair {

   public Epoch[]  read;
   protected Epoch write  = Epoch.EMPTY_EPOCH;
   protected CpuId writer = null;

   VCPair(final int size) {
      read = new Epoch[size];
      Arrays.fill(this.read, Epoch.EMPTY_EPOCH);
   }

   public void setWrite(CpuId cpu, Epoch e) {
      this.write = e;
      this.writer = cpu;
      verify();
   }

   public void setWrite(VCPair vcp) {
      this.write = vcp.write;
      this.writer = vcp.writer;
      verify();
   }

   public void clearWrite() {
      this.write = Epoch.EMPTY_EPOCH;
      this.writer = null;
   }

   public Epoch getWrite() {
      return this.write;
   }

   public CpuId getWriter() {
      return this.writer;
   }

   public boolean hasWrite() {
      verify();
      return this.writer != null;
   }

   private void verify() {
      // write and writer should always be set together
      assert !(this.writer != null && this.write.untouched());
      assert !(this.writer == null && this.write.touched());
   }

   @Override
   public String toString() {
      return "reads:" + Arrays.toString(this.read) + System.getProperty("line.separator")
             + "write:" + writer + "@" + write + System.getProperty("line.separator");
   }
}

class HardwareVCPair extends VCPair {
   public boolean[] leafRead;

   public HardwareVCPair(final int size) {
      super(size);
      leafRead = new boolean[size];
      Arrays.fill(leafRead, false);
   }

   @Override
   public String toString() {
      return super.toString() + System.getProperty("line.separator") + "leafReads:"
             + Arrays.toString(this.leafRead);
   }
}

class SoftwareVCPair extends VCPair {

   /** Whether this location was accessed via a byte-sized access. */
   public boolean byteSized = false;

   public SoftwareVCPair(final int size) {
      super(size);
   }

   @Override
   public String toString() {
      return super.toString() + System.getProperty("line.separator") + "byteSized:"
             + this.byteSized;
   }

}

public class Permissions {
   /**
    * Returns what permissions proc should have for byte mba, given the current state in hardware:
    * (cached md of other processors and in-hw status).
    * 
    * @param proc
    *           a Processor (used to lookup in-hw status)
    * @param procVc
    *           proc's vc of when it has synchronized with each other processor
    * @param mba
    *           the metadata byte to calculate permissions for
    * @param lines
    *           the set of cached lines containing mba, and the processors that own them
    * @return the MemoryAccessPermission that mba should have
    */
   public static <L extends RadishLine> MemoryAccessPermission
         usingHardwareState(Processor<L> proc, Epoch[] procVc, MetadataByteAddress mba,
                            Map<Processor<L>, L> lines) {
      HardwareVCPair vcp = Permissions.getHardwareReadWriteClocks(mba, lines, procVc.length);
      InHardwareStatus ihs = proc.inHardwareStatus.get(mba.getUnowned());
      return Permissions.usingInHwStatus(ihs, vcp, procVc);
   }

   /**
    * Just like usingHardwareState() but calculates perms for an entire cache line instead.
    */
   public static <L extends RadishLine> MemoryAccessPermission[]
         usingHardwareState(Processor<L> proc, Epoch[] procVc, final MetadataLineAddress lineAddr,
                            Map<Processor<L>, L> lines) {

      /*
       * If everything in `lines` is compact, we should calc perms for a whole data line. Otherwise,
       * calculate perms for half a data line.
       */
      boolean allCompact = true;
      for ( L l : lines.values() ) {
         allCompact &= l.isCompact();
      }
      RadishLine.Compacted compact = allCompact ? RadishLine.Compacted.COMPACT
                                               : RadishLine.Compacted.EXPANDED;

      final int DBPML = DATA_BYTES_PER_METADATA_LINE(compact);
      MemoryAccessPermission[] perms = new MemoryAccessPermission[DBPML];

      // iterate over whole md cache line
      long aligned = allCompact ? (lineAddr.get() & COMPACT_MD_ALIGNMENT_MASK()) : lineAddr.get() ; 
      MetadataByteAddress lineStart = MetadataByteAddress.Unscaled(aligned, lineAddr.owner());
      MetadataAccess a = new MetadataAccess(lineStart, DBPML);
      
      for ( ; a.keepGoing(); a.incr() ) {
         short mlo = a.metadataLineOffset(compact);
         perms[mlo] = Permissions.usingHardwareState(proc, procVc, a.addr(), lines);
      } // end byte iteration

      for ( int i = 0; i < perms.length; i++ ) {
         assert perms[i] != null;
      }

      return perms;
   }

   /**
    * Return the read/write vc's for all processors' accesses to mba
    */
   public static <L extends RadishLine> HardwareVCPair
         getHardwareReadWriteClocks(MetadataByteAddress mba, Map<Processor<L>, L> lines,
                                    int numProcs) {
      HardwareVCPair vcp = new HardwareVCPair(numProcs);

      for ( Map.Entry<Processor<L>, L> me : lines.entrySet() ) {
         L line = me.getValue();
         Processor<L> proc = me.getKey();

         // fill in read/write epochs
         ByteMetadata md = line.getLocalEpochOffset(MemoryAccessType.READ, mba);
         short pid = proc.id.get();
         vcp.read[pid] = proc.localEpochBase.add(md.epochOffset);
         vcp.leafRead[pid] = md.leafRead;

         md = line.getLocalEpochOffset(MemoryAccessType.WRITE, mba);

         if ( !vcp.hasWrite() && md.epochOffset.touched() ) {
            vcp.setWrite(proc.id, proc.localEpochBase.add(md.epochOffset));
         } else {
            // ensure there's only 1 hw writer
            assert md.epochOffset.untouched();
         }
      }

      return vcp;
   }

   /**
    * Returns what permissions proc should have, inferring a pessimistic view of the metadata in
    * software from the given InHardwareStatus.
    * 
    * @param ihs
    *           the InHardwareStatus used for sw metadata inference
    * @param rwEpochs
    *           the read/write vector clocks from all processors
    * @param procVc
    *           when proc synchronized with other processors
    */
   public static MemoryAccessPermission usingInHwStatus(InHardwareStatus ihs,
                                                        HardwareVCPair rwEpochs, Epoch[] procVc) {
      assert rwEpochs.leafRead.length == procVc.length;
      assert rwEpochs.read.length == procVc.length;

      // We add an extra vc entry for a pessimistic view of what might be in software.
      final int len = procVc.length + 1;
      final HardwareVCPair hwswVc = new HardwareVCPair(len);
      hwswVc.read = Arrays.copyOf(rwEpochs.read, len); // pads with null
      hwswVc.leafRead = Arrays.copyOf(rwEpochs.leafRead, len); // pads with false
      hwswVc.setWrite(rwEpochs);
      final Epoch[] procAndSwVc = Arrays.copyOf(procVc, len);

      final Epoch[] readVc = hwswVc.read;
      final boolean[] leafRead = hwswVc.leafRead;
      // final Epoch[] readVc = Arrays.copyOf(rwEpochs.read, len); // copyOf() pads with null
      // final Epoch[] writeVc = Arrays.copyOf(rwEpochs.write, len);
      // final boolean[] leafRead = Arrays.copyOf(rwEpochs.leafRead, len); // copyOf() pads with
      // false

      final int swProcIndex = procAndSwVc.length - 1;
      final CpuId swProc = new CpuId(swProcIndex);
      readVc[swProcIndex] = procAndSwVc[swProcIndex] = Epoch.EMPTY_EPOCH;
      final Epoch epochOne = EpochBase.INITIAL_EPOCH_BASE.add(new EpochOffset(1));
      final Epoch epochTwo = EpochBase.INITIAL_EPOCH_BASE.add(new EpochOffset(2));

      if ( ihs == null ) { // may be some reads in hardware
         // worst case: concurrent leaf read and write in sw
         // if we ha some read we ha sw write

         // read in sw is a leaf read
         leafRead[swProcIndex] = true;

         // by default, assume sw read *and* write are concurrent
         readVc[swProcIndex] = epochOne;
         hwswVc.setWrite(swProc, epochOne);
         procAndSwVc[swProcIndex] = Epoch.EMPTY_EPOCH;

         for ( int i = 0; i < readVc.length - 1; i++ ) {
            if ( readVc[i].touched() && Epoch.ha(procAndSwVc[i], readVc[i]) ) {
               // we ha a hw read, so sw write is not concurrent
               procAndSwVc[swProcIndex] = epochOne;
               // sw read is still concurrent
               readVc[swProcIndex] = epochTwo;
               break;
            }
         }

         return Permissions.calculate(hwswVc, procAndSwVc);
      }

      /*
       * int numReads = 0; int numLeafReads = 0; for ( int i = 0; i < readVc.length - 1; i++ ) { if
       * ( readVc[i].touched() ) { numReads++; if ( leafRead[i] ) { numLeafReads++; } } }
       */

      switch ( ihs ) {
      case EVERYTHING:
         // everything is in hw, so the sw md is empty
         readVc[swProcIndex] = procAndSwVc[swProcIndex] = Epoch.EMPTY_EPOCH;
         leafRead[swProcIndex] = false;
         // if ( numReads > 0 ) assert numLeafReads > 0;
         break;
      case LAST_WRITE:
         // worst case: there is a concurrent leaf read in sw
         readVc[swProcIndex] = epochOne;
         procAndSwVc[swProcIndex] = Epoch.EMPTY_EPOCH;
         leafRead[swProcIndex] = true;
         break;
      case ALL_LEAF_READS:
         /*
          * Worst case: there is a sw write that hb all (hw) reads. Thus, we ha sw write if we ha
          * some (leaf/non-leaf) read, though there may not be any (hw or sw) reads.
          */

         // if ( numReads > 0 ) assert numLeafReads > 0;

         // there shouldn't be any hw writer
         assert !hwswVc.hasWrite();

         // no reads in sw
         readVc[swProcIndex] = Epoch.EMPTY_EPOCH;
         leafRead[swProcIndex] = false;

         // by default, assume sw write is concurrent
         hwswVc.setWrite(swProc, epochOne);
         procAndSwVc[swProcIndex] = Epoch.EMPTY_EPOCH;

         for ( int i = 0; i < readVc.length - 1; i++ ) {
            if ( readVc[i].touched() && Epoch.ha(procAndSwVc[i], readVc[i]) ) {
               // sw write is not concurrent
               procAndSwVc[swProcIndex] = epochOne;
               break;
            }
         }
         break;
      }

      return Permissions.calculate(hwswVc, procAndSwVc);
   }

   /**
    * Returns what permissions thread t should have for some location l.
    * 
    * @param readVc
    *           vc of reads of l from all threads
    * @param leafRead
    *           whether the corresponding entries in readVc are leaf reads or not
    * @param writeVc
    *           vc of writes to l from all threads
    * @param threadVc
    *           vc of when t synchronized with other threads
    * @return the MemoryAccessPermission that t should have on l
    */
   public static MemoryAccessPermission calculate(HardwareVCPair hwVc, Epoch[] threadVc) {

      assert hwVc.read.length == threadVc.length;

      // should be no more than 1 writer
      boolean foundWriter = hwVc.hasWrite();
      int lastWriterIndex = foundWriter ? hwVc.getWriter().get() : -1;

      // do we happen-after all leaf reads, or do we h-a some read?

      // if we found any reads, we found >=1 leaf reads
      boolean foundReader = false;
      boolean haAllLeafReads = true;
      boolean haSomeReader = false;
      for ( int i = 0; i < hwVc.read.length; i++ ) {
         if ( hwVc.read[i].touched() ) {
            foundReader = true;
            haSomeReader |= Epoch.ha(threadVc[i], hwVc.read[i]);
            if ( hwVc.leafRead[i] ) {
               haAllLeafReads &= Epoch.ha(threadVc[i], hwVc.read[i]);
            }
         }
      }
      haAllLeafReads &= foundReader;

      boolean haLastWriter = foundWriter && Epoch.ha(threadVc[lastWriterIndex], hwVc.getWrite());

      MemoryAccessPermission perms = MemoryAccessPermission.NOTHING_OK;

      if ( foundWriter && foundReader ) {
         if ( haAllLeafReads && haLastWriter ) {
            perms = MemoryAccessPermission.WRITE_OK;
         } else if ( haSomeReader || haLastWriter ) {
            perms = MemoryAccessPermission.READ_OK;
         }

      } else if ( foundWriter && !foundReader ) {
         if ( haLastWriter ) perms = MemoryAccessPermission.WRITE_OK;

      } else if ( !foundWriter && foundReader ) {
         if ( haAllLeafReads ) perms = MemoryAccessPermission.WRITE_OK;
         else perms = MemoryAccessPermission.READ_OK;

      } else { // !foundWriter && !foundReader, e.g. no one has touched this data yet
         perms = MemoryAccessPermission.WRITE_OK;
      }

      return perms;
   }

   static class MergingResult {
      HardwareVCPair mergedClocks          = null;
      boolean        softwareHadNewerClock = false;
   }

   public static MergingResult mergeHwAndSwClocks(HardwareVCPair hwClocks, SoftwareVCPair swClocks) {
      MergingResult mr = new MergingResult();

      if ( swClocks != null ) {
         assert hwClocks.read.length == swClocks.read.length;
      }

      /*
       * Clear out stale sw entries.
       * 
       * A stale sw write w can arise when we evict w, in-hw status becomes AllLastReads, and a
       * subsequent write w' ha all readers and thus succeeds without checking sw. w is now stale.
       * 
       * We ensure that there is never >1 stale sw write, by maintaining a single-write invariant on
       * md evictions. Thus, we will see either 1) a hw and a sw write, in which case the sw write
       * is stale, or 2) only a sw write, in which case the sw write is valid.
       * 
       * Stale sw readers can never arise. A read r is only stale when a subsequent write w occurs.
       * But when w occurs, a software check will be triggered (because the not-yet-stale r is in
       * sw). w will clear out all sw md, including r, so r is never stale.
       */

      // if hw writer, clear out sw writer
      if ( hwClocks.hasWrite() && swClocks != null && swClocks.hasWrite() ) {
         swClocks.clearWrite();
      }

      // calculate up-to-date clocks by merging sw and hw
      mr.mergedClocks = new HardwareVCPair(hwClocks.read.length);

      // merge writes
      if ( swClocks == null ) {
         mr.mergedClocks.setWrite(hwClocks);

      } else {
         if ( hwClocks.hasWrite() ) {
            mr.mergedClocks.setWrite(hwClocks);
         } else if ( swClocks.hasWrite() ) {
            mr.mergedClocks.setWrite(swClocks);
         }

         if ( Epoch.Compare.gt(swClocks.getWrite(), hwClocks.getWrite()) ) {
            mr.softwareHadNewerClock = true;
         }
      }

      // merge reads
      for ( int i = 0; i < hwClocks.read.length; i++ ) {

         mr.mergedClocks.leafRead[i] = hwClocks.leafRead[i];

         if ( swClocks == null ) {
            mr.mergedClocks.read[i] = hwClocks.read[i];

         } else {
            mr.mergedClocks.read[i] = Epoch.Compare.max(hwClocks.read[i], swClocks.read[i]);

            if ( Epoch.Compare.gt(swClocks.read[i], hwClocks.read[i]) ) {
               mr.softwareHadNewerClock = true;
               // if sw has a newer read, assume (conservatively) it's a leaf read
               mr.mergedClocks.leafRead[i] = true;
            }
         }
      }

      return mr;
   }

} // end class Permissions
