package edu.washington.cs.radishsim;

import static org.junit.Assert.assertEquals;

import java.util.Deque;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

public class HierarchicalCacheTests {

   static HierarchicalCache<VILine>  l1Only        = null;
   static HierarchicalCache<VILine>  l1            = null;
   static HierarchicalCache<VILine>  l2            = null;

   static final int                  L1_ASSOC      = 2;
   static final int                  LINE_SIZE    = 4;
   static final int                  L1_CACHE_SIZE = 16;  // 4 blocks, 2 sets
   static final int                  L2_ASSOC      = 2;
   static final int                  L2_CACHE_SIZE = 16;  // 4 blocks, 2 sets

   static { MemorySystemConstants.setLineSize(LINE_SIZE); }
   
   static CacheCallbacks<VILine>     lru           = new CacheCallbacks<VILine>() {
                                                      @Override
                                                      public VILine eviction(VILine incoming, 
                                                                             Deque<VILine> set,
                                                                             CacheLevel level) {
                                                         return set.getLast();
                                                      }
                                                   };

   static CacheConfiguration<VILine> l1config      = new CacheConfiguration<VILine>() {
                                                      {
                                                         cacheSize = L1_CACHE_SIZE;
                                                         lineSize = LINE_SIZE;
                                                         assoc = L1_ASSOC;
                                                         level = CacheLevel.L1;
                                                      }
                                                   };
   static CacheConfiguration<VILine> l2config      = new CacheConfiguration<VILine>() {
                                                      {
                                                         cacheSize = L2_CACHE_SIZE;
                                                         lineSize = LINE_SIZE;
                                                         assoc = L2_ASSOC;
                                                         level = CacheLevel.L2;
                                                      }
                                                   };

   static LineFactory<VILine>        factory       = new LineFactory<VILine>() {
                                                      @Override
                                                      public VILine create() {
                                                         return new VILine();
                                                      }

                                                      @Override
                                                      public VILine create(LineAddress la, RadishLine.Compacted c) {
                                                         return new VILine(la);
                                                      }
                                                   };

   @Before
   public void setUp() {
      l1Only = new HierarchicalCache<VILine>(l1config, lru, null, factory);
      
      l2 = new HierarchicalCache<VILine>(l2config, lru, null, factory);
      l1 = new HierarchicalCache<VILine>(l1config, lru, l2, factory);
   }

   @Test
   public void test_l1access() {

      MemoryResponse<VILine> r;

      // bring 0d into the cache
      final DataByteAddress zero = new DataByteAddress(0);
      r = l1Only.request(zero);
      assertEquals(CacheLevel.MEMORY, r.whereHit);

      r = l1Only.request(zero);
      assertEquals(CacheLevel.L1, r.whereHit);

      // fetch 16d and 32d which map to set 0; this evicts 0d
      for ( int i = 1; i <= l1config.assoc; i++ ) {
         r = l1Only.request(new DataByteAddress(l1config.cacheSize * i));
         assertEquals(CacheLevel.MEMORY, r.whereHit);
      }

      // fetch 0d again, from memory
      r = l1Only.request(zero);
      assertEquals(CacheLevel.MEMORY, r.whereHit);
   }

   @Test
   public void test_l2access() {

      MemoryResponse<VILine> r;

      // bring 0d into the cache
      final DataByteAddress zero = new DataByteAddress(0);
      r = l1.request(zero);
      assertEquals(CacheLevel.MEMORY, r.whereHit);

      r = l1.request(zero);
      assertEquals(CacheLevel.L1, r.whereHit);

      // fetch 16d and 32d which map to set 0; this evicts 0d to L2
      for ( int i = 1; i <= l1config.assoc; i++ ) {
         r = l1.request(new DataByteAddress(l1config.cacheSize * i));
         assertEquals(CacheLevel.MEMORY, r.whereHit);
      }

      // fetch 0d again, from L2
      r = l1.request(zero);
      assertEquals(CacheLevel.L2, r.whereHit);

      r = l1.request(zero);
      assertEquals(CacheLevel.L1, r.whereHit);
   }

   @Test
   public void test_l2randomAccessees() {

      Random rng = new Random(123);
      for ( int i = 0; i < 100000; i++ ) {
         DataByteAddress a = new DataByteAddress(rng.nextLong());
         l1.request(a);
      }

   }

}
