package edu.washington.cs.radishsim;

import static org.junit.Assert.*;

import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import org.junit.Before;
import org.junit.Test;

public class MultiproL3Tests {
   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final CpuId                    P1                   = new CpuId(1);
   static final int                      CORES                = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_ASSOC             = 2;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 lines, 2 sets
   static final int                      L3_ASSOC             = 2;
   static final int                      L3_CACHE_SIZE        = 8;           // 2 lines, 1 set
   // (fully associative)
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    lineSize = LINE_SIZE;
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };
   static CacheConfiguration<RadishLine> l3config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    lineSize = LINE_SIZE;
                                                                    cacheSize = L3_CACHE_SIZE;
                                                                    assoc = L3_ASSOC;
                                                                    level = CacheLevel.L3;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {

         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         boolean useL3() {
            return true;
         }

         @Override
         CacheConfiguration<RadishLine> l3config() {
            return l3config;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }

         @Override
         boolean useClockGC() {
            return true; // NB!
         }
      };

      m = new Machine<RadishLine>(params);
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B

   @Test
   public void test_remoteWriteClearsOffset() {
      Processor<RadishLine> p0 = m.getProc(P0);

      for ( int j = 0; j < 20; j++ ) { // should never have >1 offset live

         for ( int i = 0; i < 2 * (1 << BITS_PER_LOCAL_EPOCH); i++ ) {
            // write
            m.cacheWrite(P0, 0, 1);

            // invalidate P0's write epoch, which resets the ref count
            m.cacheWrite(P1, 0, 1);

            // P0 does a sync op
            m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
            assertEquals(0, p0.stats.pc_EpochRollovers.get());
         }

      }
   }

   @Test
   public void test_localAccessesAllocateEpochs() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      m.cacheRead(P0, 0, 1);
      m.cacheRead(P1, 0, 1);
      assertEquals(0, p0.stats.pc_DataRaces.get());
      assertEquals(0, p1.stats.pc_DataRaces.get());

      // violates local perms
      m.cacheWrite(P1, 0, 1);

      assertEquals(0, p0.stats.pc_DataRaces.get());
      assertEquals(1, p1.stats.pc_DataRaces.get());
   }

   @Test
   public void test_trackByteRefCounts() {
      Processor<RadishLine> p0 = m.getProc(P0);

      final int limit = (1 << BITS_PER_LOCAL_EPOCH) - 1;
      for ( int i = 0; i < 2; i++ ) { // should never have >limit offsets live

         for ( int j = 0; j < limit; j++ ) {
            // read a byte
            m.cacheRead(P0, 10 + j, 1);

            // P0 does a sync op
            if ( j < limit - 1 ) {
               m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
            }

            assertEquals(0, p0.stats.pc_EpochRollovers.get());
         }

      }
   }

   @Test
   public void test_evictionRefCounts() {
      Processor<RadishLine> p0 = m.getProc(P0);

      final int BASE = 11; // maps to set 0, last byte of the line

      // allocate all offsets: 11md=22@1, 12md=24@2, 13md=26@3
      final int limit = (1 << BITS_PER_LOCAL_EPOCH) - 1;
      for ( int j = 0; j < limit; j++ ) {
         // read a byte
         m.cacheRead(P0, BASE + j, 1);

         // P0 does a sync op
         if ( j < limit - 1 ) {
            m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
         }

         assertEquals(0, p0.stats.pc_EpochRollovers.get());
      }
      // all offsets are allocated at this point, 1 in line BASE and the others in the next line

      // evict oldest offset: this is overkill since data accesses also evict things
      for ( int k = 1; k <= L1_ASSOC + L3_ASSOC; k++ ) {
         m.cacheRead(P0, BASE + (k * L1_CACHE_SIZE), 1);
         assertEquals(0, p0.stats.pc_EpochRollovers.get());
      }

      // overwrite byte with oldest offset
      m.cacheRead(P0, BASE, 1);
      assertEquals(0, p0.stats.pc_EpochRollovers.get());

      // we should have a free offset
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      assertEquals(0, p0.stats.pc_EpochRollovers.get());
      // allocate that free offset
      m.cacheRead(P0, BASE, 1);
      assertEquals(0, p0.stats.pc_EpochRollovers.get());
   }

   @Test
   public void test_evictionRefillRefCounts() {
      Processor<RadishLine> p0 = m.getProc(P0);

      final int BASE = 1; // maps to set 0, second byte of the line

      // allocate all offsets within a single line
      final int limit = (1 << BITS_PER_LOCAL_EPOCH) - 1;
      for ( int j = 0; j < limit; j++ ) {
         // read a byte
         m.cacheRead(P0, BASE + j, 1);

         // P0 does a sync op
         if ( j < limit - 1 ) {
            m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
         }

         assertEquals(0, p0.stats.pc_EpochRollovers.get());
      }
      // all offsets are allocated at this point

      // evict byte full of allocated offsets: this is overkill since data accesses also evict
      // things
      for ( int k = 1; k <= L1_ASSOC + L3_ASSOC; k++ ) {
         m.cacheRead(P0, BASE + (k * L1_CACHE_SIZE), 1);
         assertEquals(0, p0.stats.pc_EpochRollovers.get());
      }

      // fetch byte back again, with same offsets (we overwrite last byte with the same offset)
      m.cacheRead(P0, BASE + (limit - 1), 1);
      assertEquals(0, p0.stats.pc_EpochRollovers.get());
   }

   @Test
   public void test_evictionRefillRepresentableEpochs() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      // 4d maps to set 1, 4md=8 to set 0
      m.cacheRead(P0, 4, 1);
      // 0d maps to set 0, 0md=0 to set 0 as well
      // evicts 4md=8 to L3 way 0
      m.cacheRead(P0, 0, 1);

      // 8d maps to set 0, 8md=16 to set 0 as well
      // evicts 4md=8 from the L3
      m.cacheRead(P0, 8, 1);
      assertEquals(1, p0.stats.pc_MetadataEvictions.get());
      assertEquals(0, p0.stats.pc_l1md.pc_ReadHits.get());
      assertEquals(0, p0.stats.pc_l1d.pc_ReadHits.get());

      // refill 4d and 4md=8
      // epoch (1) is representable, so md is in hw now
      m.cacheRead(P0, 4, 1);
      assertEquals(InHardwareStatus.EVERYTHING,
                   p0.inHardwareStatus.get(new UnownedMetadataByteAddress(4)));

      // sync with P1
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xA);

      m.cacheWrite(P1, 4, 1);
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p0.stats.pc_DataRaces.get());
      assertEquals(0, p1.stats.pc_DataRaces.get());
   }

   @Test
   public void test_negativeEpochOffsets() {
      Processor<RadishLine> p0 = m.getProc(P0);

      for ( int i = 0; i < (1 << BITS_PER_LOCAL_EPOCH) - 1; i++ ) {
         // 0d maps to set 0, 0md=0 to set 0
         m.cacheRead(P0, i * LINE_SIZE, 1);
         assertEquals(0, p0.stats.pc_EpochRollovers.get());
         assertEquals(i + 1, p0.stats.pc_SwVcChecksTotal.get());

         m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      }
      assertEquals(1, p0.stats.pc_EpochRollovers.get());

      // evict 0md=0 to software
      // assoc/2 since d and md both map to set 0 in both caches
      for ( int i = 1; i <= (L1_ASSOC + L3_ASSOC) / 2; i++ ) {
         m.cacheRead(P0, i * L1_CACHE_SIZE, 1);
      }
      long swChecks = p0.stats.pc_SwVcChecksTotal.get();
      assertNull(p0.inHardwareStatus.get(new UnownedMetadataByteAddress(0)));

      // base epoch is now 3
      m.cacheRead(P0, 1, 1); // 1md=2,r@3, also brings in 0md=0,r@1
      assertEquals(swChecks + 1, p0.stats.pc_SwVcChecksTotal.get());
      // 1md=2,r@4 and 0md=0,r@1 should be allocated
      EpochOffset one = new EpochOffset(-2); // base + (-2) = 1
      EpochOffset four = new EpochOffset(1); // base + 1 = 4
      assertTrue(p0.epochOffsetManager.getRefCounts().contains(four));
      assertTrue(p0.epochOffsetManager.getRefCounts().contains(one));
      assertEquals(2, p0.epochOffsetManager.getRefCounts().elementSet().size());
      assertEquals(InHardwareStatus.EVERYTHING,
                   p0.inHardwareStatus.get(new UnownedMetadataByteAddress(0)));

      // p1's conflicting write to 0 can be resolved in hardware
      Processor<RadishLine> p1 = m.getProc(P1);
      m.cacheWrite(P1, 0, 1);
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
   }

   @Test
   public void test_unrepresentableEpochOffsets() {
      Processor<RadishLine> p0 = m.getProc(P0);

      for ( int i = 0; i < (1 << BITS_PER_LOCAL_EPOCH) - 1; i++ ) {
         // 0d maps to set 0, 0md=0 to set 0
         m.cacheRead(P0, i * LINE_SIZE, 1);
         assertEquals(0, p0.stats.pc_EpochRollovers.get());
         assertEquals(i + 1, p0.stats.pc_SwVcChecksTotal.get());

         m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      }
      assertEquals(1, p0.stats.pc_EpochRollovers.get());

      // base epoch is now 3

      // evict 0md=0 to software
      // assoc/2 since d and md both map to set 0 in both caches
      for ( int i = 1; i <= (L1_ASSOC + L3_ASSOC) / 2; i++ ) {
         m.cacheRead(P0, i * L1_CACHE_SIZE, 1); // r@4
      }
      assertNull(p0.inHardwareStatus.get(new UnownedMetadataByteAddress(0)));

      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      m.cacheRead(P0, 2, 1); // bring in 2md=4,r@5
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      m.cacheRead(P0, 3, 1); // bring in 3md=6,r@6

      // there aren't any free epoch offsets
      assertEquals((1 << BITS_PER_LOCAL_EPOCH) - 1, p0.epochOffsetManager.getRefCounts()
                                                                         .elementSet().size());

      long swChecks = p0.stats.pc_SwVcChecksTotal.get();
      m.cacheRead(P0, 1, 1); // 1md=2,r@6 brings in 0md=0,r@1 as well
      assertEquals(swChecks + 1, p0.stats.pc_SwVcChecksTotal.get());

      // epoch 1 (offset=-2) is not representable, and should not be allocated
      EpochOffset one = new EpochOffset(-2); // base + (-2) = 1
      assertFalse(p0.epochOffsetManager.getRefCounts().contains(one));
      assertEquals((1 << BITS_PER_LOCAL_EPOCH) - 1, p0.epochOffsetManager.getRefCounts()
                                                                         .elementSet().size());
      assertNull(p0.inHardwareStatus.get(new UnownedMetadataByteAddress(0)));

      // p1's conflicting write to 0 must go to software
      Processor<RadishLine> p1 = m.getProc(P1);
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      m.cacheWrite(P1, 0, 1);
      assertEquals(1, p1.stats.pc_SwVcChecksTotal.get());
   }

   @Test
   public void test_rolloverFlushesL3() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      // 4d maps to set 1, 4md=8 to set 0
      m.cacheRead(P0, 4, 1);
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      // 0d maps to set 0, 0md=0 to set 0 as well
      // evicts 4md=8 to L3
      m.cacheRead(P0, 0, 1);
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);

      m.cacheRead(P0, 1, 1);
      assertEquals(1, p0.stats.pc_l1d.pc_ReadHits.get());
      assertEquals(1, p0.stats.pc_l1md.pc_ReadHits.get());

      // force an epoch rollover, which flushes P0's part of the L3 (but not P1's part)
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      assertEquals(1, p0.stats.pc_EpochRollovers.get());

      // 4d maps to set 1, 4md=8 to set 0
      m.cacheRead(P1, 4, 1);
      // 0d maps to set 0, 0md=0 to set 0 as well
      // evicts P1's 4md=8 to L3, where P0's 4md=8 also resides
      m.cacheRead(P1, 0, 1);
      assertEquals(0, p1.stats.pc_l1md.pc_ReadHits.get());

      // P0's L3 md gets cleared, but P1's should remain
      assertEquals(0, p1.stats.pc_l1md.pc_ReadHits.get());
      assertEquals(2, p1.stats.pc_l1md.pc_ReadMisses.get());
      m.cacheRead(P1, 4, 1);
      assertEquals(1, p1.stats.pc_l1d.pc_ReadHits.get());
      assertEquals(1, p1.stats.pc_l3md.pc_ReadHits.get());
      assertEquals(3, p1.stats.pc_l1md.pc_ReadMisses.get());

      // P0 had 1 md hit before
      assertEquals(1, p0.stats.pc_l1d.pc_ReadHits.get());
      assertEquals(1, p0.stats.pc_l1md.pc_ReadHits.get());
      assertEquals(2, p0.stats.pc_l1md.pc_ReadMisses.get());
      m.cacheRead(P0, 4, 1);
      assertEquals(2, p0.stats.pc_l1d.pc_ReadHits.get());
      assertEquals(1, p0.stats.pc_l1md.pc_ReadHits.get());
      assertEquals(3, p0.stats.pc_l1md.pc_ReadMisses.get());
   }

   @Test
   public void test_leafReadsDowngrade() {
      Processor<RadishLine> p0 = m.getProc(P0);
      UnownedMetadataByteAddress umba0 = new UnownedMetadataByteAddress(0);

      m.cacheRead(P0, 0, 1);
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);

      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xA);
      m.cacheRead(P1, 0, 1);
      // p0's read should be a non-leaf read now

      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));

      // evict p0's read
      // assoc/2 since d and md both map to set 0 in both L1 and L3
      for ( int i = 1; i <= (L1_ASSOC + L3_ASSOC) / 2; i++ ) {
         m.cacheRead(P0, i * L1_CACHE_SIZE, 1);
      }

      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));
   }
   
   @Test
   public void test_leafReadsEvictFalseLeafRead() {
      Processor<RadishLine> p0 = m.getProc(P0);
      UnownedMetadataByteAddress umba0 = new UnownedMetadataByteAddress(0);

      m.cacheRead(P0, 0, 1);
      m.cacheRead(P1, 0, 1);
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);

      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xA);
      m.cacheRead(P1, 0, 1);
      // p1 has the leaf read now, but there's no coherence so p0 also has a (false) leaf read

      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));

      // evict p0's read
      // assoc/2 since d and md both map to set 0 in both L1 and L3
      for ( int i = 1; i <= (L1_ASSOC + L3_ASSOC) / 2; i++ ) {
         m.cacheRead(P0, i * L1_CACHE_SIZE, 1);
      }

      assertNull(p0.inHardwareStatus.get(umba0));
   }
   
   // TODO: verify that software checks do a best-effort defoliation

   @Test
   /** Random accesses without any cross-thread synchronization. */
   public void test_randomAccesses() {
      Random rng = new Random(123);

      for ( int i = 0; i < 150000; i++ ) {
         int addr = rng.nextInt() % 256;
         boolean proc = rng.nextBoolean();
         CpuId cpu = proc ? P0 : P1;
         boolean read = rng.nextBoolean();

         boolean doSyncOp = (rng.nextInt() % 10) == 0;
         if ( doSyncOp )
            m.syncOp(cpu, SynchronizationOperationType.SOURCE, false, addr);

         if ( read ) m.cacheRead(cpu, addr, 1);
         else m.cacheWrite(cpu, addr, 1);
      }
   }
   
   @Test
   /** Random accesses with actual cross-thread synchronization. */
   public void test_randomAccessesRealSync() {
      
      Random rng = new Random(1234);
      Set<Integer> sources = new HashSet<Integer>();

      for ( int i = 0; i < 150000; i++ ) {
         int addr = rng.nextInt() % 256;
         boolean proc = rng.nextBoolean();
         CpuId cpu = proc ? P0 : P1;
         boolean read = rng.nextBoolean();

         boolean doSyncOp = (rng.nextInt() % 10) == 0;
         if ( doSyncOp ) {
            if ( rng.nextBoolean() ) {
               m.syncOp(cpu, SynchronizationOperationType.SOURCE, false, addr);
               sources.add(addr);
            } else {
               m.syncOp(cpu, SynchronizationOperationType.SINK, sources.contains(addr), addr);
            }
         }
         
         int accessSize = rng.nextBoolean() ? 1 : 2 ;

         if ( read ) m.cacheRead(cpu, addr, accessSize);
         else m.cacheWrite(cpu, addr, accessSize);
      }
   }
   
}
